# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

"""
Ready-to-use Qt widgets
-----------------------

Data editors
^^^^^^^^^^^^

.. autoclass:: guidata.widgets.arrayeditor.ArrayEditor

.. autoclass:: guidata.widgets.collectionseditor.CollectionsEditor

.. autoclass:: guidata.widgets.dataframeeditor.DataFrameEditor

.. autoclass:: guidata.widgets.texteditor.TextEditor

.. autofunction:: guidata.widgets.objecteditor.oedit

Console and code editor
^^^^^^^^^^^^^^^^^^^^^^^

.. autoclass:: guidata.widgets.console.Console

.. autoclass:: guidata.widgets.console.DockableConsole

.. autoclass:: guidata.widgets.codeeditor.CodeEditor

"""
